import {
  APIEmbed,
  ActionRowBuilder,
  ButtonBuilder,
  ButtonStyle,
  ChannelResolvable,
  ChannelType,
  Client,
  ComponentType,
  EmbedBuilder,
  GatewayIntentBits,
  GuildBasedChannel,
  GuildChannelResolvable,
  GuildResolvable,
  InteractionType,
  Message,
  MessageCreateOptions,
  MessagePayload,
  PartialMessage,
  Partials,
  Typing,
  User,
  Webhook,
  WebhookMessageCreateOptions,
  WebhookType,
} from "discord.js";
import { prisma } from "./lib/prisma";
import {
  IMatrixUser,
  MatrixRoomMessage,
  MatrixRoomRedaction,
} from "./types/matrix";
import { Matrix } from "./lib/matrix";
import { FriendlyError } from "./lib/utils";
import { Turndown } from "./lib/turndown";
import { Glue } from "./lib/glue";

class Discord_ {
  client = new Client({
    intents: [
      GatewayIntentBits.Guilds,
      GatewayIntentBits.GuildMembers,
      GatewayIntentBits.GuildPresences,
      GatewayIntentBits.GuildMessages,
      GatewayIntentBits.MessageContent,
      GatewayIntentBits.GuildMessageTyping,
      GatewayIntentBits.GuildMessageReactions,
    ],
    partials: [Partials.Message],
  });

  constructor() {
    this.client.on("messageCreate", this.handleMessageCreate.bind(this));
    this.client.on("messageUpdate", this.handleMessageUpdate.bind(this));
    this.client.on("messageDelete", this.handleMessageDelete.bind(this));
    this.client.on("typingStart", this.handleTypingStart.bind(this));

    this.client.on("ready", () => {
      console.log(`Logged in as ${this.client.user?.tag}`);
    });
  }

  login(token: string) {
    this.client.login(token);
  }

  async getAttachmentURL(
    guild_id: string,
    channel_id: string,
    message_id: string,
    attachment_id: string
  ) {
    const guild = await this.client.guilds.fetch(guild_id);
    const channel = await guild.channels.fetch(channel_id);
    if (!channel || channel.type !== ChannelType.GuildText) return;
    // force the messages to be fetched from scratch to avoid caching invalid attachment URLs
    const messages = await channel.messages.fetch({
      around: message_id,
      cache: false,
    });

    const message = messages.first();
    if (!message) {
      return undefined;
    }

    return message.attachments.get(attachment_id)?.url;
  }

  async handleMessageCreate(message: Message) {
    if (message.author.id === this.client.user!.id) return; // ignore bot's own messages
    if (message.webhookId) return; // if it's a webhook we don't care about it
    if (!message.guildId) return; // we only care about guild messages

    const bridged = await Glue.getBridges({
      discord_guild: message.guildId,
      discord_channel: message.channelId,
    });
    if (!bridged.length) return; // no valid bridges

    var content = message.content;
    content = await this.replaceRoleMentions(message.guildId, content);

    let matrixClient;
    try {
      matrixClient = await Matrix.for(`_discord_${message.author.id}`);
      await matrixClient.syncProfile(
        message.author.displayName,
        `mxc://${process.env.HOST}/${Buffer.from(["avatar", message.author.id].join(",")).toString("base64url")}`
      );
    } catch (e) {
      console.error("Failed to get bridged ghost user", e);
      return;
    }

    for (const room of bridged) {
      try {
        await Matrix.ensureUserInRoom(
          room.matrix_room_id,
          matrixClient.username
        );

        const reference = await message.fetchReference().catch((e) => null);
        const inReplyTo_eventID =
          reference &&
          ((await prisma.discordMessages.findFirst({
            where: {
              discord_id: reference.id,
              discord_channel_id: reference.channelId,
            },
          })) ||
            (await prisma.matrixMessages.findFirst({
              where: {
                discord_id: reference.id,
                discord_channel_id: reference.channelId,
              },
            })));

        await matrixClient
          .sendEvent<MatrixRoomMessage>(
            room.matrix_room_id,
            {
              type: "m.room.message",
              content: {
                msgtype: "m.text",
                body: content,
                "m.mentions": {
                  user_ids: Matrix.scanForMatrixMentions(message.content),
                },
                "m.relates_to": inReplyTo_eventID
                  ? {
                      "m.in_reply_to": {
                        event_id: inReplyTo_eventID.matrix_event_id,
                      },
                    }
                  : undefined,
              },
            },
            [room.matrix_room_id, message.id].join(",")
          )
          .then((result) => {
            if (result.status === 200) {
              return prisma.discordMessages.create({
                data: {
                  discord_id: message.id,
                  discord_guild_id: message.guildId!,
                  discord_channel_id: message.channelId,
                  discord_author_id: message.author.id,
                  matrix_event_id: result.data.event_id,
                  matrix_room_id: room.matrix_room_id,
                },
              });
            } else {
              console.error("Failed to send message", result);
            }
          });

        for (const [id, attachment] of message.attachments) {
          let attachmentType: MatrixRoomMessage["content"]["msgtype"] =
            "m.file";

          if (attachment.contentType?.startsWith("image/")) {
            attachmentType = "m.image";
          }
          if (attachment.contentType?.startsWith("video/")) {
            attachmentType = "m.video";
          }
          if (attachment.contentType?.startsWith("audio/")) {
            attachmentType = "m.audio";
          }

          let content: MatrixRoomMessage["content"] = {
            msgtype: attachmentType,
            body: "attachment." + attachment.name.split(".").at(-1),
            url: `mxc://${process.env.HOST}/${Buffer.from(["attachment", message.guildId, message.channelId, message.id, id].join(",")).toString("base64url")}`,
            info: {
              mimetype: attachment.contentType!,
              h: attachment.height || undefined,
              w: attachment.width || undefined,
            },
            "m.mentions": {
              user_ids: [],
            },
          } as any; // no idea why the type isn't valid

          await matrixClient
            .sendEvent<MatrixRoomMessage>(
              room.matrix_room_id,
              {
                type: "m.room.message",
                content,
              },
              [room.matrix_room_id, message.id, id].join(",")
            )
            .then((result) => {
              if (result.status === 200) {
                return prisma.discordMessages.create({
                  data: {
                    discord_id: message.id,
                    discord_guild_id: message.guildId!,
                    discord_channel_id: message.channelId,
                    discord_author_id: message.author.id,
                    attachment_id: id,
                    matrix_event_id: result.data.event_id,
                    matrix_room_id: room.matrix_room_id,
                  },
                });
              } else {
                console.error("Failed to send message", result);
              }
            });
        }
      } catch (e) {
        if (e instanceof FriendlyError) {
          console.error("Failed to bridge to room [friendly]", room, e);
        } else {
          console.error("Failed to bridge to room", room, e);
        }
      }
    }
  }

  async handleMessageUpdate(
    oldMessage: Message | PartialMessage,
    newMessage: Message | PartialMessage
  ) {
    if (newMessage.webhookId) return;
    if (!newMessage.author || !newMessage.guildId || !oldMessage.guildId)
      return;

    const messages = await prisma.discordMessages.findMany({
      where: {
        discord_id: newMessage.id,
      },
    });

    if (!messages.length) return;

    const toDeleteAttachments = new Set<string>();

    for (const [id] of oldMessage.attachments) {
      // loop through old message attachment IDs
      // if the old ID isn't in the new message, it's been deleted
      if (!newMessage.attachments.get(id)) {
        toDeleteAttachments.add(id);
      }
    }

    const matrixClient = await this.matrix(newMessage.author);
    await matrixClient.syncProfile(newMessage.author.displayName);

    for (const attachmentId of toDeleteAttachments) {
      const dbMessages = messages.filter(
        (m) => m.attachment_id === attachmentId
      );

      for (const dbMessage of dbMessages) {
        try {
          await Matrix.ensureUserInRoom(
            dbMessage.matrix_room_id,
            matrixClient.username
          );
          await matrixClient.redactEvent(
            dbMessage.matrix_room_id,
            dbMessage.matrix_event_id,
            dbMessage.id + "_" + Date.now()
          );
          await prisma.discordMessages.delete({
            where: {
              id: dbMessage.id,
            },
          });
        } catch (e) {
          console.log("Failed to delete event", dbMessage, e);
        }
      }
    }

    let content = newMessage.content;

    const contentMessages = messages.filter((m) => !m.attachment_id);
    for (const dbMessage of contentMessages) {
      await Matrix.ensureUserInRoom(
        dbMessage.matrix_room_id,
        matrixClient.username
      );
      const resp = await matrixClient.sendEvent<MatrixRoomMessage>(
        dbMessage.matrix_room_id,
        {
          type: "m.room.message",
          content: {
            msgtype: "m.text",
            body: "* " + content,
            "m.relates_to": {
              rel_type: "m.replace",
              event_id: dbMessage.matrix_event_id,
            },
            "m.new_content": {
              msgtype: "m.text",
              body: content!,
            },
            "m.mentions": {
              // we shouldn't mention anyone in a edit event
              user_ids: [],
            },
          },
        },
        [
          dbMessage.matrix_room_id,
          newMessage.id,
          newMessage.editedTimestamp,
        ].join("_")
      );
    }

    // if (resp.status === 200) {
    //   await prisma.discordMessages.update({
    //     where: { discord_id: newMessage.id },
    //     data: {
    //       discord_id: newMessage.id,
    //       matrix_event_id: resp.data.event_id,
    //     },
    //   });
    // } else {
    //   await newMessage.reply(
    //     "Failed to send editted message: " +
    //       (resp.data.error || resp.data.errcode)
    //   );
    // }
  }

  async handleMessageDelete(message: Message | PartialMessage) {
    if (!message.author || !message.guildId) return;

    const discordMessages = await prisma.discordMessages.findMany({
      where: { discord_id: message.id },
    });
    const user_matrixClient = await this.matrix(message.author);
    const system_matrixClient = await Matrix.for("_discord_bot");

    for (const dbMessage of discordMessages) {
      try {
        await Matrix.ensureUserInRoom(
          dbMessage.matrix_room_id,
          user_matrixClient.username
        );
        await user_matrixClient.redactEvent(
          dbMessage.matrix_room_id,
          dbMessage.matrix_event_id,
          message.id + "_" + Date.now()
        );
        await prisma.discordMessages.delete({
          where: {
            id: dbMessage.id,
          },
        });
      } catch (e) {
        console.error("Failed to delete message", dbMessage, e);
      }
    }

    const matrixMessages = await prisma.matrixMessages.findMany({
      where: { discord_id: message.id },
    });

    for (const dbMessage of matrixMessages) {
      try {
        await system_matrixClient.redactEvent(
          dbMessage.matrix_room_id,
          dbMessage.matrix_event_id,
          message.id + "_" + Date.now()
        );
      } catch (e) {
        console.error("Failed to delete message", dbMessage, e);
      }
    }

    // if (resp.status === 200) {
    //   await prisma.discordMessages.delete({
    //     where: { discord_id: message.id },
    //   });
    // } else {
    //   await message.channel.send(
    //     "<@" +
    //       message.author.id +
    //       "> failed to send delete: " +
    //       (resp.data.error || resp.data.errcode)
    //   );
    // }
  }

  async handleTypingStart(typing: Typing) {
    if (typing.user.id === this.client.user!.id) return; // ignore bot's own messages
    if (typing.user.bot) return; // if it's a webhook we don't care about it
    if (!typing.inGuild()) return;

    const bridges = await Glue.getBridges({
      discord_guild: typing.guild.id,
      discord_channel: typing.channel.id,
    });
    const user = await Matrix.for(`_discord_${typing.user.id}`);

    for (const bridge of bridges) {
      await Matrix.ensureUserInRoom(
        bridge.matrix_room_id,
        `_discord_${typing.user.id}`
      );
      await user.sendTyping(bridge.matrix_room_id);
    }
  }

  async sendMessage(
    guild_id: string,
    channel_id: string,
    user: IMatrixUser,
    content: string,
    options: {
      replyTo?: {
        /**
         * Discord Message ID
         */
        messageId: string;
        /**
         * Display name for a user replying to
         */
        who: string;
      };
      /**
       * IMPORTANT: attachment URLs are fetched and streamed to discord
       */
      attachment?: { url: string; name: string };
    } = {}
  ) {
    const webhook = await this.getWebhookFor(guild_id, channel_id);

    let displayName = user.displayname
      ? user.displayname + " (" + user.userId + ")"
      : user.userId;

    if (displayName.length > 80) {
      displayName = user.userId;
    }

    displayName = displayName.slice(0, 80);

    let components = [];

    if (options.replyTo) {
      /**
       * Is the replyTo a Discord message?
       * If the replyTo user starts with an @, it's mentioning a Matrix ID
       */
      const isDiscord = !options.replyTo.who.startsWith("@");
      const replyToWho = isDiscord
        ? "@" + (await this.client.users.fetch(options.replyTo.who)).username
        : options.replyTo.who;

      components.push(
        new ActionRowBuilder<ButtonBuilder>().addComponents(
          new ButtonBuilder()
            .setStyle(ButtonStyle.Link)
            .setURL(
              `https://discord.com/channels/${guild_id}/${channel_id}/${options.replyTo.messageId}`
            )
            .setLabel("In Reply To"),
          new ButtonBuilder()
            .setStyle(ButtonStyle.Secondary)
            .setCustomId("dummy")
            .setDisabled(true)
            .setLabel(replyToWho)
        )
      );
    }

    let messageContent = {
      username: displayName,
      avatarURL: Matrix.getMediaURL(user.avatar_url),
      content,
      components,
      allowedMentions: {
        parse: ["users"],
        roles: [],
      } as const,
      embeds: [] as APIEmbed[],
    };

    let messageId: string | undefined;

    let files: { attachment: Buffer; name: string }[] = [];

    if (options.attachment) {
      messageId = (
        await webhook.send({
          ...messageContent,
          content: "*Uploading attachment*",
        })
      ).id;

      try {
        const request = await fetch(options.attachment.url);
        if (request.status > 399)
          throw new Error("Unknown status: " + request.status);

        const stream = await request.arrayBuffer();

        files.push({
          attachment: Buffer.from(stream),
          name: options.attachment.name,
        });
      } catch (e) {
        console.error(
          "Failed to upload attachment",
          guild_id,
          channel_id,
          user,
          e
        );
        messageContent.embeds.push(
          new EmbedBuilder()
            .setColor("Red")
            .setDescription("Failed to upload attachment")
            .toJSON()
        );
      }
    }

    if (messageId) {
      return webhook.editMessage(messageId, {
        ...messageContent,
        files,
      });
    }

    return webhook.send({
      username: displayName,
      avatarURL: Matrix.getMediaURL(user.avatar_url),
      content,
      components,
      files,
      allowedMentions: {
        parse: ["users"],
        roles: [],
      },
    });
  }

  async editMessage(
    guild_id: string,
    channel_id: string,
    message_id: string,
    content: string
  ) {
    const webhook = await this.getWebhookFor(guild_id, channel_id);
    return await webhook.editMessage(message_id, {
      content,
      allowedMentions: {
        parse: ["users"],
        roles: [],
      },
    });
  }

  async deleteMessage(
    guild_id: string,
    channel_id: string,
    message_id: string,
    isWebhook = true
  ) {
    if (isWebhook) {
      const webhook = await this.getWebhookFor(guild_id, channel_id);
      return await webhook.deleteMessage(message_id);
    } else {
      const { channel } = await this.getGuildChannel(guild_id, channel_id);
      return await channel.messages.fetch(message_id).then((msg) => {
        return msg.delete();
      });
    }
  }

  async getGuildChannel(guild_id: string, channel_id: string) {
    const guild = await this.client.guilds.fetch(guild_id);
    if (!guild) throw new Error("unknown_guild " + guild_id);

    const channel = await guild.channels.fetch(channel_id);
    if (!channel) throw new Error("unknown_channel");
    if (channel.type !== ChannelType.GuildText)
      throw new Error("channel_is_not_text");

    return { guild, channel };
  }

  async getWebhookFor(guild_id: string, channel_id: string) {
    const { channel } = await this.getGuildChannel(guild_id, channel_id);

    const db_webhook = await prisma.discordWebhooks.findFirst({
      where: { channel_id },
    });
    let webhook: Webhook<WebhookType.Incoming>;

    if (db_webhook) {
      webhook = (await this.client.fetchWebhook(
        db_webhook.webhook_id,
        db_webhook.webhook_token
      )) as Webhook<WebhookType.Incoming>;
    } else {
      webhook = await channel.createWebhook({
        name: "matrix-bridge",
      });
      await prisma.discordWebhooks.create({
        data: {
          channel_id: channel.id,
          webhook_id: webhook.id,
          webhook_token: webhook.token!,
        },
      });
    }

    return webhook;
  }

  async matrix(user: User) {
    return await Matrix.for(`_discord_${user.id}`);
  }

  async replaceRoleMentions(guild_id: string, content: string) {
    const guild = await this.client.guilds.fetch(guild_id);
    if (!guild) return content;

    const ROLE_REGEX = /<@&[0-9]+>/gi;

    const matches = new Set(
      [...content.matchAll(ROLE_REGEX)]
        .map((match) => match?.[0])
        .filter((a) => a)
    );

    for (const match of matches) {
      const role_id = match.replace("<@&", "").replace(">", "");
      const role = await guild.roles.fetch(role_id);
      if (!role) continue;

      content = content.replace(new RegExp(match, "g"), "@" + role.name);
    }

    return content;
  }

  /**
   * Convert Matrix org.matrix.custom.html to Discord
   * @param message
   */
  matrixToDiscord(message: MatrixRoomMessage) {
    if (!("format" in message.content)) return message.content.body;

    let content = message.content.formatted_body;

    // const MATRIX_USER = Matrix.getMatrixUserRegex();
    // const matrixMention = new RegExp(
    //   `<a href="https://matrix\.to/#/(${MATRIX_USER.source})">.*(?!<a)</a>`,
    //   "gi"
    // );

    // const matches = new Set([...content.matchAll(matrixMention)]);

    // for (const [mention, userId] of matches) {
    //   if (!userId.startsWith("@_discord_")) continue;

    //   const discordId = userId
    //     .replace("@_discord_", "")
    //     .replace(":" + process.env.MATRIX_HOMESERVER, "");

    //   content = content.replace(
    //     new RegExp(mention, "g"),
    //     "<@" + discordId + ">"
    //   );
    // }

    content = Turndown.turndown(content);
    content = content.replace(/\bhttps?:\/\/matrix\.to\/[^ )]*/g, "<$&>");

    return content;
  }
}

export const Discord = new Discord_();
