import EventEmitter from "eventemitter3";
import {
  IClientEvent,
  IMatrixUser,
  MXC,
  MatrixRoomMembership,
  SentClientEvent,
} from "../types/matrix";
import { FriendlyError } from "./utils";

const { MATRIX_AS_TOKEN } = process.env;
/**
 * Homeserver friendly name (not direct access)
 */
const MATRIX_HOMESERVER =
  process.env.MATRIX_HOMESERVER_HOST ?? process.env.MATRIX_HOMESERVER;

interface MatrixEvents {
  joined: (user_id: IUsername) => void;
}

class EventHolder extends EventEmitter<MatrixEvents> {}
const events = new EventHolder();

export class Matrix {
  static instances = new Map<IUsername, Matrix>();

  readonly username: IUsername;
  private access_token: string;
  private expires_at: Date;
  private refresh_token: string;

  constructor(
    username: IUsername,
    { access_token, expires_in_ms, refresh_token }: AuthSession
  ) {
    this.username = username;
    this.access_token = access_token;
    this.expires_at = new Date(Date.now() + expires_in_ms);
    this.refresh_token = refresh_token;
  }

  private fetch: (typeof Matrix)["fetch"] = (_endpoint, ...args) => {
    const [endpoint, rawParams] = _endpoint.split("?");

    const params = new URLSearchParams(rawParams);
    params.set("user_id", "@" + this.username + ":" + MATRIX_HOMESERVER);

    return Matrix.fetch((endpoint + "?" + params) as any, ...args);
  };

  /**
   * Get homeserver access
   *
   * * internal: used by this bridge internally
   * * external: passed to external service (like discord or client)
   *
   * @param use
   * @returns
   */
  private static getHomeserver(use: "internal" | "external") {
    switch (use) {
      case "internal": {
        let protocol =
          process.env.NODE_ENV === "development" ? "http" : "https";
        return protocol + "://" + process.env.MATRIX_HOMESERVER;
      }
      case "external": {
        return "https://" + MATRIX_HOMESERVER;
      }
    }
  }

  private static async fetch<T = unknown>(
    endpoint: `/_matrix${string}`,
    method = "GET",
    data?: any
  ): Promise<
    | { status: 200; data: T }
    | { status: 400 | 403 | 429; data: { errcode: string; error?: string } }
  > {
    console.log("[->Matrix] " + method + " " + endpoint);

    const req = await fetch(`${this.getHomeserver("internal")}${endpoint}`, {
      method,
      headers: {
        "Content-Type": "application/json",
        Authorization: "Bearer " + MATRIX_AS_TOKEN,
      },
      body: JSON.stringify(data),
    });

    const res = await req.json();

    return {
      status: req.status as any,
      data: res,
    };
  }

  async sendTyping(roomId: string) {
    return this.fetch<{}>(
      `/_matrix/client/v3/rooms/${roomId}/typing/@${this.username}:${MATRIX_HOMESERVER}`,
      "PUT",
      {
        timeout: 1000 * 10,
        typing: true,
      }
    );
  }

  /**
   * Send event to the server
   *
   * - events (state or not) can be sent to HS can include ?ts= to change timestamp
   *
   * @see https://spec.matrix.org/v1.10/application-service-api/#timestamp-massaging
   */
  async sendEvent<Event extends IClientEvent = IClientEvent>(
    roomId: string,
    event: SentClientEvent<Event>,
    txnId: string
  ) {
    return this.fetch<{ event_id: string }>(
      `/_matrix/client/v3/rooms/${roomId}/send/${event.type}/${txnId}`,
      "PUT",
      event.content
    );
  }

  async redactEvent(roomId: string, eventId: string, txnId: string) {
    return this.fetch<{ event_id: string }>(
      `/_matrix/client/v3/rooms/${roomId}/redact/${eventId}/${txnId}`,
      "PUT",
      {}
    );
  }

  async inviteUser(roomId: string, user: IUsername) {
    return await this.fetch<{}>(
      `/_matrix/client/v3/rooms/${roomId}/invite`,
      "POST",
      {
        user_id: `@${user}:${MATRIX_HOMESERVER}`,
      }
    );
  }

  static async isUserInRoom(roomId: string, user: IUsername): Promise<boolean> {
    const req = await this.fetch<MatrixRoomMembership["content"]>(
      `/_matrix/client/v3/rooms/${roomId}/state/m.room.member/@${user}:${MATRIX_HOMESERVER}`
    );

    if ("errcode" in req.data) {
      // if(req.data.errcode === "M_NOT_FOUND")
      return false;
    } else {
      return req.data.membership === "join";
    }
  }

  /**
   * Use the main account to make sure the mentioned user is in the room, if not, invite
   * @param roomId
   * @param user
   */
  static async ensureUserInRoom(roomId: string, user: IUsername) {
    const req = await this.fetch<MatrixRoomMembership["content"]>(
      `/_matrix/client/v3/rooms/${roomId}/state/m.room.member/@${user}:${MATRIX_HOMESERVER}`
    );
    const superMatrixClient = await Matrix.for("_discord_bot");
    const userMatrixClient = await Matrix.for(user);

    const waitForJoin = () => {
      return new Promise<void>((res) => {
        const handle = (user_id: IUsername) => {
          if (user_id !== user) return;

          res();
          events.off("joined", handle);
        };

        events.on("joined", handle);
      });
    };

    if ("errcode" in req.data) {
      switch (req.data.errcode) {
        case "M_NOT_FOUND":
          // invite
          const resp = await superMatrixClient.inviteUser(roomId, user);
          if (resp.status !== 200) {
            throw new FriendlyError(resp.data.error || resp.data.errcode);
          }

          await waitForJoin();
          break;
      }
    } else {
      switch (req.data.membership) {
        case "invite":
          // make user join
          await userMatrixClient.joinRoom(roomId);
          break;
        case "ban":
          // throw error
          throw new Error("banned");
        case "join":
          // user already here
          // no-op
          break;
        case "leave":
        case "knock": {
          // invite
          const resp = await superMatrixClient.inviteUser(roomId, user);
          if (resp.status !== 200) {
            throw new FriendlyError(resp.data.error || resp.data.errcode);
          }

          await waitForJoin();
        }
      }
    }
  }

  async syncProfile(displayname?: string, avatar_url?: string) {
    const fqun = `@${this.username}:${MATRIX_HOMESERVER}`;
    const profile = await Matrix.getUserProfile(fqun as any);

    if (profile.displayname !== displayname) {
      await this.fetch(
        `/_matrix/client/v3/profile/${fqun}/displayname`,
        "PUT",
        {
          displayname,
        }
      );
    }

    if (profile.avatar_url !== avatar_url) {
      await this.fetch(`/_matrix/client/v3/profile/${fqun}/avatar_url`, "PUT", {
        avatar_url,
      });
    }
  }

  /**
   * Gets a Matrix instance for a specific username
   * @param username
   */
  static async for(username: IUsername): Promise<Matrix> {
    if (!this.instances.has(username)) {
      let session: AuthSession;
      const registerReq = await this.registerUser(username);

      if (registerReq === "already_registered") {
        session = await this.getAccessToken(username);
      } else {
        session = registerReq;
      }

      const instance = new Matrix(username, session);

      this.instances.set(username, instance);

      return instance;
    } else {
      return this.instances.get(username)!;
    }
  }

  private static async registerUser(
    username: IUsername
  ): Promise<"already_registered" | AuthSession> {
    // POST /_matrix/client/v3/register
    // {
    //   type: "m.login.application_service",
    //   username
    // }

    const register = await this.fetch<{
      access_token: string;
      expires_in_ms: number;
      refresh_token: string;
    }>(`/_matrix/client/v3/register`, "POST", {
      type: "m.login.application_service",
      username,
    });

    if (register.status !== 200) {
      if (register.data.errcode === "M_USER_IN_USE") {
        return "already_registered";
      } else {
        console.error("unknown error code while registering user", register);
        return "already_registered";
      }
    } else {
      return register.data;
    }
  }

  /**
   * Get access token for username
   * @param username
   */
  private static async getAccessToken(
    username: IUsername
  ): Promise<AuthSession> {
    // POST /_matrix/client/v3/login
    // {
    //   type: "m.login.application_service",
    //   identifier: {
    //     type: "m.id.user",
    //     user: username
    //   }
    // }

    const loginReq = await this.fetch<{
      access_token: string;
      expires_in_ms: number;
      refresh_token: string;
    }>(`/_matrix/client/v3/login`, "POST", {
      type: "m.login.application_service",
      identifier: {
        type: "m.id.user",
        user: username,
      },
    });

    if (loginReq.status !== 200) {
      console.error("unknown error code while getting access_token", loginReq);
      return {} as any;
    } else {
      return loginReq.data;
    }
  }

  /**
   * Sync events
   *
   * - MUST include virtual user (?user_id)
   * - this shouldn't be used, but instead use events pushed to the service
   *
   * @see https://spec.matrix.org/v1.10/application-service-api/#using-sync-and-events
   */
  sync() {}

  async joinRoom(roomIdOrAlias: string) {
    const req = await this.fetch(
      `/_matrix/client/v3/join/${roomIdOrAlias}`,
      "POST",
      {}
    );

    events.emit("joined", this.username);

    return req;
  }

  async leaveRoom(roomId: string) {
    const req = await this.fetch(
      `/_matrix/client/v3/rooms/${roomId}/leave`,
      "POST",
      {}
    );

    return req;
  }

  /**
   * Get all local aliases for room_id
   *
   * TODO: caching
   *
   * @param room_id
   * @returns
   */
  static async getAliasesForRoom(room_id: string) {
    const req = await this.fetch<{ aliases: string[] }>(
      `/_matrix/client/v3/rooms/${room_id}/aliases`,
      "GET"
    );

    if (req.status === 200) {
      return req.data.aliases.filter(
        (a) =>
          a.startsWith("#_discord_") &&
          (a.indexOf(":") > -1 ? a.endsWith(":" + MATRIX_HOMESERVER) : true)
      );
    } else {
      return [];
    }
  }

  /**
   * populate matrix user profile from user_id
   *
   * TODO: cache
   *
   * @param user_id
   */
  static async getUserProfile(
    user_id: `@${string}:${string}`
  ): Promise<IMatrixUser> {
    const req = await this.fetch<{
      avatar_url?: MXC;
      displayname?: string;
    }>(`/_matrix/client/v3/profile/${user_id}`, "GET");

    if (req.status !== 200) {
      // silently fail, this isn't neccessary information anyway
      return {
        userId: user_id,
      };
    }

    return {
      userId: user_id,
      displayname: req.data.displayname,
      avatar_url: req.data.avatar_url,
    };
  }

  /**
   * Convert mxc:// url to download url
   * @param mxc
   * @returns
   */
  static getMediaURL(mxc?: MXC, use: "internal" | "external" = "external") {
    if (!mxc) return undefined;

    return mxc.replace(
      "mxc://",
      `${this.getHomeserver(use)}/_matrix/media/v3/download/`
    );
  }

  static getMatrixUserRegex() {
    const localpart_regex = /[a-z0-9\._=\-\/\+]{1,100}/;

    // https://stackoverflow.com/a/26987741
    const domainpart_regex =
      /(((?!-))(xn--|_)?[a-z0-9-]{0,61}[a-z0-9]{1,1}\.)*(xn--)?([a-z0-9][a-z0-9\-]{0,60}|[a-z0-9-]{1,30}\.[a-z]{2,})/;

    return new RegExp(
      "@" + localpart_regex.source + ":" + domainpart_regex.source,
      "gi"
    );
  }

  /**
   * Return matrix user mentions inside text
   *
   * @note this does not follow the spec entirely
   *
   * @param content
   */
  static scanForMatrixMentions(content: string): IMatrixUser["userId"][] {
    const MATRIX_USER = new RegExp(
      this.getMatrixUserRegex().source + "(\\w|$)",
      "gi"
    );

    return [...content.matchAll(MATRIX_USER)]
      .map((match) => match?.[0])
      .filter((a) => a) as any;
  }

  static async sendAlivePing() {
    return await this.fetch(
      `/_matrix/client/v1/appservice/${process.env.MATRIX_AS_ID}/ping`,
      "POST",
      {}
    );
  }
}

type IUsername = `_discord_${string}`;
type AuthSession = {
  access_token: string;
  expires_in_ms: number;
  refresh_token: string;
};
