// this file contains types for the matrix spec, not the matrix handler

export type MatrixError = {
  errcode: string;
  error: string;
};

export type SentClientEvent<Event extends IClientEvent> = Omit<
  Event,
  "event_id" | "origin_server_ts" | "room_id" | "sender"
>;

export interface IClientEvent<Content = unknown> {
  event_id: string;
  type: string;

  content: Content;
  prev_content?: Content;
  origin_server_ts: number;
  room_id: string;
  sender: IMatrixUser["userId"];

  state_key?: string;
  unsigned?: MatrixUnsignedData<Content>;
}

export interface IEphemeralEvent<
  Type extends string = string,
  Content = unknown,
> {
  type: Type;
  content: Content;
}

export type MatrixEphemeralEvent = MatrixEphemeralTyping;

export type MatrixEphemeralTyping = IEphemeralEvent<
  "m.typing",
  {
    user_ids: string[];
  }
> & {
  room_id: string;
};

export type MatrixUnsignedData<Content = unknown> = {
  age?: number;
  prev_content?: Content;
  redacted_because?: IClientEvent;
  transaction_id?: string;
};

export type MXC = `mxc://${string}/${string}`;

export interface IMatrixUser {
  userId: `@${string}:${string}`;
  avatar_url?: MXC;
  displayname?: string;
}

export type MatrixRoomMembership = {
  type: "m.room.membership";
} & IClientEvent<{
  displayname?: string;
  membership: "invite" | "join" | "knock" | "leave" | "ban";
}>;

export type MatrixRoomRedaction = {
  type: "m.room.redaction";
} & IClientEvent<{ reason?: string; redacts: string }>;

/**
 * @template IsEdit if this type is apart of an edit {@see MessageEdit}
 */
export type MatrixRoomMessage<IsEdit extends boolean = false> = {
  type: "m.room.message";
} & IClientEvent<
  (
    | FormattableBody<"m.text">
    | FormattableBody<"m.emote">
    | FormattableBody<"m.notice">
    | (FormattableBody<"m.image"> & HasAttachment<ImageInfo>)
    | (FormattableBody<"m.file"> & HasAttachment<FileInfo>)
    | (FormattableBody<"m.audio"> & HasAttachment<AudioInfo>)
    | {
        msgtype: "m.location";
        body: string;
        geo_uri: string;
        info: LocationInfo;
      }
    | (FormattableBody<"m.video"> & HasAttachment<VideoInfo>)
  ) &
    (IsEdit extends false
      ? /**
         * We want to strictly enforce mentions so avoid unintentional mentions
         *
         * @since Matrix Spec v1.7
         */
        MessageMention & (MessageEdit | MessageReply | {})
      : {})
>;

type MessageEdit = {
  "m.relates_to": {
    rel_type: "m.replace";
    event_id: string;
  };
  "m.new_content": MatrixRoomMessage<true>["content"];
};

type MessageReply = {
  "m.relates_to": {
    "m.in_reply_to": {
      event_id: string;
    };
  };
};

type MessageMention = {
  "m.mentions": {
    user_ids: IMatrixUser["userId"][];
    /**
     * Mention @room
     */
    room?: boolean;
  };
};

type FormattableBody<Type extends string> =
  | { msgtype: Type; body: string }
  | {
      msgtype: Type;
      body: string;
      format: "org.matrix.custom.html";
      formatted_body: string;
    };

type HasAttachment<Info> = { filename?: string; url: string; info: Info };

type ImageInfo = {
  h?: number;
  w?: number;
  mimetype?: string;
  size?: number; // bytes
  thumbnail_url?: string;
};

type FileInfo = {
  mimetype?: string;
  size?: number; // bytes
  thumbnail_url?: string;
};

type AudioInfo = {
  mimetype?: string;
  duration?: string; // milliseconds
  size?: number; // bytes
};

type LocationInfo = {
  thumbnail_url?: string;
};

type VideoInfo = {
  duration?: number; // milliseconds
  h?: number;
  w?: number;
  mimetype?: string;
  size?: number; // bytes
  thumbnail_url?: string;
};
